# Author: Nathan Trouvain at 15/02/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Dict, Iterator, Sequence, Union
from uuid import uuid4

import numpy as np

from .type import MappedData, Shape
from .utils import progress
from .utils.validation import check_vector, is_mapping


def _distant_model_inputs(model):
    """Get inputs for distant Nodes in a Model used as feedback or teacher.
    These inputs should be already computed by other Nodes."""
    input_data = {}
    for p, c in model.edges:
        if p in model.input_nodes:
            input_data[c.name] = p.state_proxy()
    return input_data


def _remove_input_for_feedback(node) -> Union["Node", "Model"]:
    """Remove inputs nodes from feedback Model and gather remaining nodes
    into a new Model. Allow getting inputs for feedback model from its input
    nodes states."""
    from .model import Model

    all_nodes = set(node.nodes)
    input_nodes = set(node.input_nodes)
    filtered_nodes = list(all_nodes - input_nodes)
    filtered_edges = [edge for edge in node.edges if edge[0] not in input_nodes]

    # return a single Node if Model - Inputs = Node
    # else return Model - Inputs = Reduced Model
    if len(filtered_nodes) == 1:
        return list(filtered_nodes)[0]
    return Model(filtered_nodes, filtered_edges, name=str(uuid4()))


def check_one_sequence(
    x: Union[np.ndarray, Sequence[np.ndarray]],
    expected_dim=None,
    caller=None,
    allow_timespans=True,
):
    caller_name = caller.name + " is" if caller is not None else ""

    if expected_dim is not None and not hasattr(expected_dim, "__iter__"):
        expected_dim = (expected_dim,)

    x_new = check_vector(
        x, allow_reshape=True, allow_timespans=allow_timespans, caller=caller
    )
    data_dim = x_new.shape[1:]

    # Check x dimension
    if expected_dim is not None:
        if len(expected_dim) != len(data_dim):
            raise ValueError(
                f"{caller_name} expecting {len(expected_dim)} inputs "
                f"but received {len(data_dim)}: {x_new}."
            )
        for dim in expected_dim:
            if all([dim != ddim for ddim in data_dim]):
                raise ValueError(
                    f"{caller_name} expecting data of shape "
                    f"{expected_dim} but received shape {data_dim}."
                )
    return x_new


# expected_dim = ((m, n), o, (p, q, r), ...)
def check_n_sequences(
    x,
    expected_dim=None,
    allow_n_sequences=True,
    allow_n_inputs=True,
    allow_timespans=True,
    caller=None,
):
    if expected_dim is not None:
        if not hasattr(expected_dim, "__iter__"):
            expected_dim = (expected_dim,)
        n_inputs = len(expected_dim)

        # I
        if n_inputs > 1:
            if isinstance(x, (list, tuple)):
                x_new = [x[i] for i in range(len(x))]
                timesteps = []
                for i in range(n_inputs):
                    dim = (expected_dim[i],)
                    x_new[i] = check_n_sequences(
                        x[i],
                        expected_dim=dim,
                        caller=caller,
                        allow_n_sequences=allow_n_sequences,
                        allow_timespans=allow_timespans,
                        allow_n_inputs=allow_n_inputs,
                    )
                    if isinstance(x_new[i], (list, tuple)):
                        timesteps.append(tuple([x_.shape[0] for x_ in x_new[i]]))
                    else:
                        dim = dim[0]
                        if not hasattr(dim, "__len__"):
                            dim = (dim,)
                        if len(dim) + 2 > len(x_new[i].shape) >= len(dim) + 1:
                            timesteps.append((x_new[i].shape[0],))
                        else:
                            timesteps.append((x_new[i].shape[1],))

                if len(np.unique([len(t) for t in timesteps])) > 1 or any(
                    [
                        len(np.unique([t[i] for t in timesteps])) > 1
                        for i in range(len(timesteps[0]))
                    ]
                ):
                    raise ValueError("Inputs with different timesteps")
            else:
                raise ValueError("Expecting several inputs.")
        else:  # L
            dim = expected_dim[0]
            if not hasattr(dim, "__len__"):
                dim = (dim,)

            if isinstance(x, (list, tuple)):
                if not allow_n_sequences:
                    raise TypeError("No lists, only arrays.")
                x_new = [x[i] for i in range(len(x))]
                for i in range(len(x)):
                    x_new[i] = check_one_sequence(
                        x[i],
                        allow_timespans=allow_timespans,
                        expected_dim=dim,
                        caller=caller,
                    )
            else:
                if len(x.shape) <= len(dim) + 1:  # only one sequence
                    x_new = check_one_sequence(
                        x,
                        expected_dim=dim,
                        allow_timespans=allow_timespans,
                        caller=caller,
                    )
                elif len(x.shape) == len(dim) + 2:  # several sequences
                    # if not allow_n_sequences:
                    #     raise TypeError("No lists, only arrays.")
                    x_new = x
                    for i in range(len(x)):
                        x_new[i] = check_one_sequence(
                            x[i],
                            allow_timespans=allow_timespans,
                            expected_dim=dim,
                            caller=caller,
                        )
                else:  # pragma: no cover
                    x_new = check_vector(
                        x,
                        allow_reshape=True,
                        allow_timespans=allow_timespans,
                        caller=caller,
                    )
    else:
        if isinstance(x, (list, tuple)):
            x_new = [x[i] for i in range(len(x))]
            for i in range(len(x)):
                if allow_n_inputs:
                    x_new[i] = check_n_sequences(
                        x[i],
                        allow_n_sequences=allow_n_sequences,
                        allow_timespans=allow_timespans,
                        allow_n_inputs=False,
                        caller=caller,
                    )
                elif allow_n_sequences:
                    x_new[i] = check_n_sequences(
                        x[i],
                        allow_n_sequences=False,
                        allow_timespans=allow_timespans,
                        allow_n_inputs=False,
                        caller=caller,
                    )
                else:
                    raise TypeError("No lists, only arrays.")
        else:
            x_new = check_one_sequence(
                x, allow_timespans=allow_timespans, caller=caller
            )

    return x_new


def _check_node_io(
    x,
    receiver_nodes=None,
    expected_dim=None,
    caller=None,
    io_type="input",
    allow_n_sequences=True,
    allow_n_inputs=True,
    allow_timespans=True,
):
    noteacher_msg = f"Nodes can not be used as {io_type}" + " for {}."
    notonline_msg = "{} is not trained online."

    x_new = None
    # Caller is a Model
    if receiver_nodes is not None:
        if not is_mapping(x):
            x_new = {n.name: x for n in receiver_nodes}
        else:
            x_new = x.copy()

        for node in receiver_nodes:
            if node.name not in x_new:
                # Maybe don't fit nodes a second time
                if io_type == "target" and node.fitted:
                    continue
                else:
                    raise ValueError(f"Missing {io_type} data for node {node.name}.")

            if (
                callable(x_new[node.name])
                and hasattr(x_new[node.name], "initialize")
                and hasattr(x_new[node.name], "is_initialized")
                and hasattr(x_new[node.name], "output_dim")
            ):
                if io_type == "target":
                    if node.is_trained_online:
                        register_teacher(
                            node,
                            x_new.pop(node.name),
                            expected_dim=node.output_dim,
                        )
                    else:
                        raise TypeError(
                            (noteacher_msg + notonline_msg).format(node.name, node.name)
                        )
                else:
                    raise TypeError(noteacher_msg.format(node.name))
            else:
                if io_type == "target":
                    dim = node.output_dim
                else:
                    dim = node.input_dim

                x_new[node.name] = check_n_sequences(
                    x_new[node.name],
                    expected_dim=dim,
                    caller=node,
                    allow_n_sequences=allow_n_sequences,
                    allow_n_inputs=allow_n_inputs,
                    allow_timespans=allow_timespans,
                )
    # Caller is a Node
    else:
        if (
            callable(x)
            and hasattr(x, "initialize")
            and hasattr(x, "is_initialized")
            and hasattr(x, "output_dim")
        ):
            if io_type == "target":
                if caller.is_trained_online:
                    register_teacher(
                        caller,
                        x,
                        expected_dim=expected_dim,
                    )
                else:
                    raise TypeError(
                        (noteacher_msg + notonline_msg).format(caller.name, caller.name)
                    )
            else:
                raise TypeError(noteacher_msg.format(caller.name))
        else:
            x_new = check_n_sequences(
                x,
                expected_dim=expected_dim,
                caller=caller,
                allow_n_sequences=allow_n_sequences,
                allow_n_inputs=allow_n_inputs,
                allow_timespans=allow_timespans,
            )

    # All x are teacher nodes, no data to return
    if is_mapping(x_new) and io_type == "target" and len(x_new) == 0:
        return None

    return x_new


def register_teacher(caller, teacher, expected_dim=None):
    target_dim = None
    if teacher.is_initialized:
        target_dim = teacher.output_dim

    if (
        expected_dim is not None
        and target_dim is not None
        and expected_dim != target_dim
    ):
        raise ValueError()

    caller._teacher = DistantFeedback(
        sender=teacher, receiver=caller, callback_type="teacher"
    )


def check_xy(
    caller,
    x,
    y=None,
    input_dim=None,
    output_dim=None,
    allow_n_sequences=True,
    allow_n_inputs=True,
    allow_timespans=True,
):
    """Prepare one step of input and target data for a Node or a Model.

    Preparation may include:
        - reshaping data to ([inputs], [sequences], timesteps, features);
        - converting non-array objects to array objects;
        - checking if n_features is equal to node input or output dimension.

    This works on numerical data and teacher nodes.

    Parameters
    ----------
    caller: Node or Model
        Node or Model requesting inputs/targets preparation.
    x : array-like of shape ([inputs], [sequences], timesteps, features)
        Input array or sequence of input arrays containing a single timestep of
        data.
    y : array-like of shape ([sequences], timesteps, features) or Node, optional
        Target array containing a single timestep of data, or teacher Node or
        Model
        yielding target values.
    input_dim, output_dim : int or tuple of ints, optional
        Expected input and target dimensions, if available.

    Returns
    -------
    array-like of shape ([inputs], 1, n), array-like of shape (1, n) or Node
        Processed input and target vectors.
    """

    if input_dim is None and hasattr(caller, "input_dim"):
        input_dim = caller.input_dim

    # caller is a Model
    if hasattr(caller, "input_nodes"):
        input_nodes = caller.input_nodes
    # caller is a Node
    else:
        input_nodes = None

    x_new = _check_node_io(
        x,
        receiver_nodes=input_nodes,
        expected_dim=input_dim,
        caller=caller,
        io_type="input",
        allow_n_sequences=allow_n_sequences,
        allow_n_inputs=allow_n_inputs,
        allow_timespans=allow_timespans,
    )

    y_new = y
    if y is not None:
        # caller is a Model
        if hasattr(caller, "trainable_nodes"):
            output_dim = None
            trainable_nodes = caller.trainable_nodes

        # caller is a Node
        else:
            trainable_nodes = None
            if output_dim is None and hasattr(caller, "output_dim"):
                output_dim = caller.output_dim

        y_new = _check_node_io(
            y,
            receiver_nodes=trainable_nodes,
            expected_dim=output_dim,
            caller=caller,
            io_type="target",
            allow_n_sequences=allow_n_sequences,
            allow_timespans=allow_timespans,
            allow_n_inputs=False,
        )

    return x_new, y_new


class DistantFeedback:
    def __init__(self, sender, receiver, callback_type="feedback"):
        self._sender = sender
        self._receiver = receiver
        self._callback_type = callback_type

        # used to store a reduced version of the feedback if needed
        # when feedback is a Model (inputs of the feedback Model are suppressed
        # in the reduced version, as we do not need then to re-run them
        # because we assume they have already run during the forward call)
        self._reduced_sender = None

        self._clamped = False
        self._clamped_value = None

    def __call__(self):
        if not self.is_initialized:
            self.initialize()
        return self.call_distant_node()

    @property
    def is_initialized(self):
        return self._sender.is_initialized

    @property
    def output_dim(self):
        return self._sender.output_dim

    @property
    def name(self):
        return self._sender.name

    def call_distant_node(self):
        """Call a distant Model for feedback or teaching
        (no need to run the input nodes again)"""
        if self._clamped:
            self._clamped = False
            return self._clamped_value

        if self._reduced_sender is not None:
            if len(np.unique([n._fb_flag for n in self._sender.nodes])) > 1:
                input_data = _distant_model_inputs(self._sender)

                if hasattr(self._reduced_sender, "nodes"):
                    return self._reduced_sender.call(input_data)
                else:
                    reduced_name = self._reduced_sender.name
                    return self._reduced_sender.call(input_data[reduced_name])
            else:
                fb_outputs = [n.state() for n in self._sender.output_nodes]
                if len(fb_outputs) > 1:
                    return fb_outputs
                else:
                    return fb_outputs[0]
        else:
            return self._sender.state_proxy()

    def initialize(self):
        """Initialize a distant Model or Node (used as feedback sender or teacher)."""
        msg = f"Impossible to get {self._callback_type} "
        msg += "from {} for {}: {} is not initialized or has no input/output_dim"

        reduced_model = None
        if hasattr(self._sender, "input_nodes"):
            for n in self._sender.input_nodes:
                if not n.is_initialized:
                    try:
                        n.initialize()
                    except RuntimeError:
                        raise RuntimeError(
                            msg.format(
                                self._sender.name,
                                self._receiver.name,
                                self._sender.name,
                            )
                        )

            input_data = _distant_model_inputs(self._sender)
            reduced_model = _remove_input_for_feedback(self._sender)

            if not reduced_model.is_initialized:
                if hasattr(reduced_model, "nodes"):
                    reduced_model.initialize(x=input_data)
                else:
                    reduced_name = reduced_model.name
                    reduced_model.initialize(x=input_data[reduced_name])
                self._sender._is_initialized = True
        else:
            try:
                self._sender.initialize()
            except RuntimeError:  # raise more specific error
                raise RuntimeError(
                    msg.format(
                        self._sender.name, self._receiver.name, self._sender.name
                    )
                )

        self._reduced_sender = reduced_model

    def zero_feedback(self):
        """A null feedback vector. Returns None if the Node receives
        no feedback."""
        if hasattr(self._sender, "output_nodes"):
            zeros = []
            for output in self._sender.output_nodes:
                zeros.append(output.zero_state())
            if len(zeros) == 1:
                return zeros[0]
            else:
                return zeros
        else:
            return self._sender.zero_state()

    def clamp(self, value):
        self._clamped_value = check_n_sequences(
            value,
            expected_dim=self._sender.output_dim,
            caller=self._sender,
            allow_n_sequences=False,
        )
        self._clamped = True


def call(node, x, from_state=None, stateful=True, reset=False):
    """One-step call, without input check."""
    with node.with_state(from_state, stateful=stateful, reset=reset):
        state = node._forward(node, x)
        node._state = state.astype(node.dtype)
        node._flag_feedback()

    return state


def train(
    node,
    X,
    Y=None,
    call_node=True,
    force_teachers=True,
    learn_every=1,
    from_state=None,
    stateful=True,
    reset=False,
):
    seq_len = X.shape[0]
    seq = (
        progress(range(seq_len), f"Training {node.name}")
        if seq_len > 1
        else range(seq_len)
    )

    with node.with_state(from_state, stateful=stateful, reset=reset):
        states = np.zeros((seq_len, node.output_dim))
        for i in seq:
            x = np.atleast_2d(X[i, :])

            y = None
            if node._teacher is not None:
                y = node._teacher()
            elif Y is not None:
                y = np.atleast_2d(Y[i, :])

            if call_node:
                s = call(node, x)
            else:
                s = node.state()

            if force_teachers:
                node.set_state_proxy(y)

            if i % learn_every == 0 or seq_len == 1:
                node._train(node, x=x, y=y)

            states[i, :] = s

    return states


class _Node(ABC):
    """Node base class for type checking and interface inheritance."""

    _factory_id = -1
    _registry = list()
    _name: str

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._factory_id = -1
        cls._registry = list()

    def __repr__(self):
        klas = type(self).__name__
        hypers = [(str(k), str(v)) for k, v in self._hypers.items()]
        all_params = ["=".join((k, v)) for k, v in hypers]
        all_params += [f"in={self.input_dim}", f"out={self.output_dim}"]
        return f"'{self.name}': {klas}(" + ", ".join(all_params) + ")"

    def __setstate__(self, state):
        curr_name = state.get("_name")
        if curr_name in type(self)._registry:
            new_name = curr_name + "-(copy)"
            state["_name"] = new_name
        self.__dict__ = state

    def __del__(self):
        try:
            type(self)._registry.remove(self._name)
        except (ValueError, AttributeError):
            pass

    def __getattr__(self, item):
        if item in ["_params", "_hypers"]:
            raise AttributeError()
        if item in self._params:
            return self._params.get(item)
        elif item in self._hypers:
            return self._hypers.get(item)
        else:
            raise AttributeError(f"'{str(item)}'")

    def __setattr__(self, name, value):
        if hasattr(self, "_params") and name in self._params:
            self._params[name] = value
        elif hasattr(self, "_hypers") and name in self._hypers:
            self._hypers[name] = value
        else:
            super(_Node, self).__setattr__(name, value)

    def __call__(self, *args, **kwargs) -> np.ndarray:
        return self.call(*args, **kwargs)

    def __rshift__(self, other: Union["_Node", Sequence["_Node"]]) -> "Model":
        from .ops import link

        return link(self, other)

    def __rrshift__(self, other: Union["_Node", Sequence["_Node"]]) -> "Model":
        from .ops import link

        return link(other, self)

    def __and__(self, other: Union["_Node", Sequence["_Node"]]) -> "Model":
        from .ops import merge

        return merge(self, other)

    def _get_name(self, name=None):
        if name is None:
            type(self)._factory_id += 1
            _id = self._factory_id
            name = f"{type(self).__name__}-{_id}"

        if name in type(self)._registry:
            raise NameError(
                f"Name '{name}' is already taken "
                f"by another node. Node names should "
                f"be unique."
            )

        type(self)._registry.append(name)
        return name

    @property
    def name(self) -> str:
        """Name of the Node or Model."""
        return self._name

    @name.setter
    def name(self, value):
        type(self)._registry.remove(self.name)
        self._name = self._get_name(value)

    @property
    def params(self) -> Dict[str, Any]:
        """Parameters of the Node or Model."""
        return self._params

    @property
    def hypers(self) -> Dict[str, Any]:
        """Hyperparameters of the Node or Model."""
        return self._hypers

    @property
    def is_initialized(self) -> bool:
        return self._is_initialized

    @property
    @abstractmethod
    def input_dim(self) -> Shape:
        raise NotImplementedError()

    @property
    @abstractmethod
    def output_dim(self) -> Shape:
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_trained_offline(self) -> bool:
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_trained_online(self) -> bool:
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_trainable(self) -> bool:
        raise NotImplementedError()

    @property
    @abstractmethod
    def fitted(self) -> bool:
        raise NotImplementedError()

    @is_trainable.setter
    @abstractmethod
    def is_trainable(self, value: bool):
        raise NotImplementedError()

    def get_param(self, name: str) -> Any:
        if name in self._params:
            return self._params.get(name)
        elif name in self._hypers:
            return self._hypers.get(name)
        else:
            raise NameError(f"No parameter named '{name}' found in node {self}")

    @abstractmethod
    def copy(
        self, name: str = None, copy_feedback: bool = False, shallow: bool = False
    ) -> "_Node":
        raise NotImplementedError()

    @abstractmethod
    def initialize(self, x: MappedData = None, y: MappedData = None):
        raise NotImplementedError()

    @abstractmethod
    def reset(self, to_state: np.ndarray = None) -> "_Node":
        raise NotImplementedError()

    @contextmanager
    @abstractmethod
    def with_state(self, state=None, stateful=False, reset=False) -> Iterator["_Node"]:
        raise NotImplementedError()

    @contextmanager
    @abstractmethod
    def with_feedback(
        self, feedback=None, stateful=False, reset=False
    ) -> Iterator["_Node"]:
        raise NotImplementedError()



"""
=======================================================
Experimental features (:mod:`reservoirpy.experimental`)
=======================================================

.. currentmodule:: reservoirpy.experimental

You can find in this module all features in development and testing for
future releases.

.. warning::

    All features in the experimental module may still be under heavy
    development. You can still report any bug by opening an issue
    as explained in :ref:`opening_an_issue`.

Nodes
=====

.. autosummary::
    :toctree: generated/
    :template: autosummary/class.rst

    LIF - Leaky Integrate and Fire RNN (Liquid State Machine)
    Add - Add two vectors.
    BatchFORCE - Fast implementation of FORCE algorithm.
    RandomChoice - Randomly select features in a vector of data.
    AsabukiNorm - Normalization as defined in Asabuki et al. (2018)
"""

# Author: Nathan Trouvain at 03/02/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from .add import Add
from .batchforce import BatchFORCE
from .norm import AsabukiNorm
from .randomchoice import RandomChoice
from .sklearn import from_sklearn
from .spiking.lif import LIF

__all__ = [
    "Add",
    "BatchFORCE",
    "AsabukiNorm",
    "RandomChoice",
    "from_sklearn",
    "LIF",
]



# Author: Nathan Trouvain at 27/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from copy import deepcopy
from multiprocessing import Manager

import numpy as np
from joblib import Parallel, delayed

from .._base import _Node, call
from ..model import FrozenModel
from ..utils import _obj_from_kwargs, progress, verbosity
from ..utils.graphflow import dispatch
from ..utils.model_utils import to_data_mapping
from ..utils.parallel import get_joblib_backend
from ..utils.validation import is_mapping
from .io import Input
from .readouts import Ridge
from .reservoirs import NVAR, Reservoir

_LEARNING_METHODS = {"ridge": Ridge}

_RES_METHODS = {"reservoir": Reservoir, "nvar": NVAR}


def _run_partial_fit_fn(esn, x, y, lock, warmup):
    # the 'loky' and 'multiprocessing' backends already deep-copies the ESN. See
    # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
    _esn = deepcopy(esn)
    _esn.reservoir.reset()

    original_readout_name = (
        esn.readout.name[:-7]
        if esn.readout.name.endswith("-(copy)")
        else esn.readout.name
    )
    original_reservoir_name = (
        esn.reservoir.name[:-7]
        if esn.reservoir.name.endswith("-(copy)")
        else esn.reservoir.name
    )

    seq_len = len(x[list(x)[0]])
    states = np.zeros((seq_len, esn.reservoir.output_dim))

    for i, (x, forced_feedback, _) in enumerate(dispatch(x, y, shift_fb=True)):
        with _esn.readout.with_feedback(forced_feedback[original_readout_name]):
            states[i, :] = call(_esn.reservoir, x[original_reservoir_name])

    esn.readout.partial_fit(states, y[original_readout_name], warmup=warmup, lock=lock)

    return states[-1]


def _run_fn(
    esn, idx, x, forced_fb, return_states, from_state, stateful, reset, shift_fb
):
    _esn = deepcopy(esn)

    original_reservoir_name = (
        esn.reservoir.name[:-7]
        if esn.reservoir.name.endswith("-(copy)")
        else esn.reservoir.name
    )

    X = {_esn.reservoir.name: x[original_reservoir_name]}

    states = _allocate_returned_states(_esn, X, return_states)

    with _esn.with_state(from_state, stateful=stateful, reset=reset):
        for i, (x_step, forced_feedback, _) in enumerate(
            dispatch(X, forced_fb, shift_fb=shift_fb)
        ):
            _esn._load_proxys()
            with _esn.with_feedback(forced_feedback):
                state = _esn._call(x_step, return_states=return_states)

            if is_mapping(state):
                for name, value in state.items():
                    states[name][i, :] = value
            else:
                states["readout"][i, :] = state

    _esn._clean_proxys()
    return idx, states


def _allocate_returned_states(model, inputs, return_states=None):
    """Create empty placeholders for model outputs."""
    seq_len = inputs[list(inputs.keys())[0]].shape[0]
    vulgar_names = {"reservoir": model.reservoir, "readout": model.readout}

    # pre-allocate states
    if return_states == "all":
        states = {
            name: np.zeros((seq_len, n.output_dim)) for name, n in vulgar_names.items()
        }
    elif isinstance(return_states, (list, tuple)):
        states = {
            name: np.zeros((seq_len, n.output_dim))
            for name, n in {name: vulgar_names[name] for name in return_states}.items()
        }
    else:
        states = {"readout": np.zeros((seq_len, model.readout.output_dim))}

    return states


def _sort_and_unpack(states, return_states=None):
    """Maintain input order (even with parallelization on)"""
    states = sorted(states, key=lambda s: s[0])
    states = {n: [s[1][n] for s in states] for n in states[0][1].keys()}

    for n, s in states.items():
        if len(s) == 1:
            states[n] = s[0]

    if len(states) == 1 and return_states is None:
        states = states["readout"]

    return states


class ESN(FrozenModel):
    """Echo State Networks as a Node, with parallelization of state update.

    This Node is provided as a wrapper for reservoir and readout nodes. Execution
    is distributed over several workers when:

    - the ``workers`` parameters is equal to `n>1` (using `n` workers) or
      `n<=-1` (using `max_available_workers - n` workers)
    - Several independent sequences of inputs are fed to the model at runtime.

    When parallelization is enabled, internal states of the reservoir will be reset
    to 0 at the beginning of every independent sequence of inputs.

    Note
    ----
        This node can not be connected to other nodes. It is only provided as a
        convenience Node to speed up processing of large datasets with "vanilla"
        Echo State Networks.

    :py:attr:`ESN.params` **list**:

    ================== =================================================================
    ``reservoir``      A :py:class:`~reservoirpy.nodes.Reservoir` or a :py:class:`~reservoirpy.nodes.NVAR` instance.
    ``readout``        A :py:class:`~reservoirpy.nodes.Ridge` instance.
    ================== =================================================================

    :py:attr:`ESN.hypers` **list**:

    ==================== ===============================================================
    ``workers``          Number of workers for parallelization (1 by default).
    ``backend``          :py:mod:`joblib` backend to use for parallelization  (``loky`` by default,).
    ``reservoir_method`` Type of reservoir, may be "reservoir" or "nvar" ("reservoir" by default).
    ``learning_method``  Type of readout, by default "ridge".
    ``feedback``         Is readout connected to reservoir through feedback (False by default).
    ==================== ===============================================================

    Parameters
    ----------
    reservoir_method : {"reservoir", "nvar"}, default to "reservoir"
        Type of reservoir, either a :py:class:`~reservoirpy.nodes.Reservoir` or
        a :py:class:`~reservoirpy.nodes.NVAR`.
    learning_method : {"ridge"}, default to "ridge"
        Type of readout. The only method supporting parallelization for now is the
        :py:class:`~reservoirpy.nodes.Ridge` readout.
    reservoir : Node, optional
        A Node instance to use as a reservoir,
        such as a :py:class:`~reservoirpy.nodes.Reservoir` node.
    readout : Node, optional
        A Node instance to use as a readout,
        such as a :py:class:`~reservoirpy.nodes.Ridge` node
        (only this one is supported).
    feedback : bool, defaults to False
        If True, the readout is connected to the reservoir through
        a feedback connection.
    use_raw_inputs : bool, defaults to False
        If True, the input is directly fed to the readout. See
        :ref:`/user_guide/advanced_demo.ipynb#Input-to-readout-connections`.
    Win_bias : bool, default to True
        If True, add an input bias to the reservoir.
    Wout_bias : bool, default to True
        If True, add a bias term to the reservoir states entering the readout.
    workers : int, default to 1
        Number of workers used for parallelization. If set to -1, all available workers
        (threads or processes) are used.
    backend : a :py:mod:`joblib` backend, default to "loky"
        A parallelization backend.
    name : str, optional
        Node name.

    See Also
    --------
    Reservoir
    Ridge
    NVAR

    Example
    -------
    >>> from reservoirpy.nodes import Reservoir, Ridge, ESN
    >>> reservoir, readout = Reservoir(100, sr=0.9), Ridge(ridge=1e-6)
    >>> model = ESN(reservoir=reservoir, readout=readout, workers=-1)
    """

    def __init__(
        self,
        reservoir_method="reservoir",
        learning_method="ridge",
        reservoir: _Node = None,
        readout: _Node = None,
        feedback=False,
        Win_bias=True,
        Wout_bias=True,
        workers=1,
        backend=None,
        name=None,
        use_raw_inputs=False,
        **kwargs,
    ):
        msg = "'{}' is not a valid method. Available methods for {} are {}."

        if reservoir is None:
            if reservoir_method not in _RES_METHODS:
                raise ValueError(
                    msg.format(reservoir_method, "reservoir", list(_RES_METHODS.keys()))
                )
            else:
                klas = _RES_METHODS[reservoir_method]
                kwargs["input_bias"] = Win_bias
                reservoir = _obj_from_kwargs(klas, kwargs)

        if readout is None:
            if learning_method not in _LEARNING_METHODS:
                raise ValueError(
                    msg.format(
                        learning_method, "readout", list(_LEARNING_METHODS.keys())
                    )
                )
            else:
                klas = _LEARNING_METHODS[learning_method]
                kwargs["input_bias"] = Wout_bias
                readout = _obj_from_kwargs(klas, kwargs)

        if feedback:
            reservoir <<= readout

        if use_raw_inputs:
            source = Input()
            super(ESN, self).__init__(
                nodes=[reservoir, readout, source],
                edges=[(source, reservoir), (reservoir, readout), (source, readout)],
                name=name,
            )
        else:
            super(ESN, self).__init__(
                nodes=[reservoir, readout], edges=[(reservoir, readout)], name=name
            )

        self._hypers.update(
            {
                "workers": workers,
                "backend": backend,
                "reservoir_method": reservoir_method,
                "learning_method": learning_method,
                "feedback": feedback,
            }
        )

        self._params.update({"reservoir": reservoir, "readout": readout})

        self._trainable = True
        self._is_fb_initialized = False

    @property
    def is_trained_offline(self) -> bool:
        return True

    @property
    def is_trained_online(self) -> bool:
        return False

    @property
    def is_fb_initialized(self):
        return self._is_fb_initialized

    @property
    def has_feedback(self):
        """Always returns False, ESNs are not supposed to receive external
        feedback. Feedback between reservoir and readout must be defined
        at ESN creation."""
        return False

    def _call(self, x=None, return_states=None, *args, **kwargs):
        data = x[self.reservoir.name]

        state = call(self.reservoir, data)
        call(self.readout, state)

        state = {}
        if return_states == "all":
            for node in ["reservoir", "readout"]:
                state[node] = getattr(self, node).state()
        elif isinstance(return_states, (list, tuple)):
            for name in return_states:
                state[name] = getattr(self, name).state()
        else:
            state = self.readout.state()

        return state

    def state(self, which="reservoir"):
        if which == "reservoir":
            return self.reservoir.state()
        elif which == "readout":
            return self.readout.state()
        else:
            raise ValueError(
                f"'which' parameter of {self.name} "
                f"'state' function must be "
                f"one of 'reservoir' or 'readout'."
            )

    def run(
        self,
        X=None,
        forced_feedbacks=None,
        from_state=None,
        stateful=True,
        reset=False,
        shift_fb=True,
        return_states=None,
    ):
        X, forced_feedbacks = to_data_mapping(self, X, forced_feedbacks)

        self._initialize_on_sequence(X[0], forced_feedbacks[0])

        backend = get_joblib_backend(workers=self.workers, backend=self.backend)

        seq = progress(X, f"Running {self.name}")

        with self.with_state(from_state, reset=reset, stateful=stateful):
            with Parallel(n_jobs=self.workers, backend=backend) as parallel:
                states = parallel(
                    delayed(_run_fn)(
                        self,
                        idx,
                        x,
                        y,
                        return_states,
                        from_state,
                        stateful,
                        reset,
                        shift_fb,
                    )
                    for idx, (x, y) in enumerate(zip(seq, forced_feedbacks))
                )

        return _sort_and_unpack(states, return_states=return_states)

    def fit(
        self, X=None, Y=None, warmup=0, from_state=None, stateful=True, reset=False
    ):
        X, Y = to_data_mapping(self, X, Y)
        self._initialize_on_sequence(X[0], Y[0])

        self.initialize_buffers()

        if (self.workers > 1 or self.workers < 0) and self.backend != "sequential":
            lock = Manager().Lock()
        else:
            lock = None

        backend = get_joblib_backend(workers=self.workers, backend=self.backend)

        seq = progress(X, f"Running {self.name}")
        with self.with_state(from_state, reset=reset, stateful=stateful):
            with Parallel(n_jobs=self.workers, backend=backend) as parallel:
                last_states = parallel(
                    delayed(_run_partial_fit_fn)(self, x, y, lock, warmup)
                    for x, y in zip(seq, Y)
                )

            if verbosity():  # pragma: no cover
                print(f"Fitting node {self.name}...")

            # setting the reservoir state from the last timeseries
            self.reservoir._state = last_states[-1]
            self.readout.fit()

        return self



# Author: Nathan Trouvain at 23/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import os
import shutil
from contextlib import contextmanager

import numpy as np
import pytest
from scipy import sparse

from ... import activationsfunc as F
from .. import ESN, load_compat


@contextmanager
def create_old_model(
    sparse_W=True, input_bias=True, feedback=False, train=False, fbfunc=F.identity
):

    parent_dir = os.path.dirname(os.path.realpath(__file__))

    d = parent_dir + "dummy_model"

    if os.path.exists(d):
        shutil.rmtree(d)

    in_dim = 10 + int(input_bias)
    wfb = None

    if sparse_W:
        w = sparse.random(100, 100, format="csr", dtype=np.float64)
    else:
        w = np.random.normal(0, 0.1, size=(100, 100))

    win = np.random.normal(0, 0.1, size=(100, in_dim))
    if feedback:
        wfb = np.random.normal(0, 0.1, size=(100, 10))

    esn = ESN(
        lr=0.3, W=w, Win=win, input_bias=input_bias, Wfb=wfb, ridge=1e-8, fbfunc=fbfunc
    )

    if train:
        X, Y = [np.ones((100, 10))], [np.ones((100, 10))]
        esn.train(X, Y)

    esn.save(d)

    yield d, esn

    if os.path.exists(d):
        shutil.rmtree(d)


@pytest.mark.parametrize(
    "sparse,bias,feedback,train,fbfunc",
    (
        (True, False, False, False, F.identity),
        (True, False, False, True, F.identity),
        (True, True, False, False, F.identity),
        (True, False, True, True, F.sigmoid),
        (True, True, True, False, F.tanh),
        (False, True, False, False, F.identity),
        (False, False, True, True, F.softmax),
    ),
)
def test_load_files_from_v2(sparse, bias, feedback, fbfunc, train):
    with create_old_model(
        sparse_W=sparse, input_bias=bias, feedback=feedback, train=train, fbfunc=fbfunc
    ) as m:

        dirname, esn = m[0], m[1]

        esn2 = load_compat(dirname)

        X = np.ones((100, 10))

        if train is False and feedback is False:
            with pytest.raises(RuntimeError):
                esn2.run(X)

            esn2 = load_compat(dirname)
            esn2.fit(X, X)
            res = esn2.run(X)
            assert res.shape == (100, 10)
        else:
            res = esn2.run(X)
            assert res.shape == (100, 10)



# Author: Nathan Trouvain at 15/02/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_equal

from .._base import DistantFeedback, check_n_sequences, check_one_sequence, check_xy
from .dummy_nodes import *


def idfn(val):
    if isinstance(val, np.ndarray):
        return str(val.shape)
    if isinstance(val, list):
        return f"list[{len(val)}]"
    if isinstance(val, dict):
        return str(val)
    else:
        return val


@pytest.mark.parametrize(
    "x,kwargs,expects",
    [
        (np.ones((1, 5)), {}, np.ones((1, 5))),
        (np.ones((5,)), {}, np.ones((1, 5))),
        (np.ones((1, 5)), {"expected_dim": 6}, ValueError),
        ("foo", {}, TypeError),
        (1, {}, np.ones((1, 1))),
        ([np.ones((1, 5)), np.ones((1, 6))], {}, TypeError),
        (np.ones((5, 5)), {}, np.ones((5, 5))),
        (np.ones((2, 5)), {}, np.ones((2, 5))),
        (np.ones((2, 5)), {"expected_dim": 5}, np.ones((2, 5))),
        (np.ones((2, 5)), {"expected_dim": 2}, ValueError),
        (np.ones((2, 5, 2)), {"expected_dim": 2}, ValueError),
        (np.ones((2, 5, 2)), {"expected_dim": (5, 2)}, np.ones((2, 5, 2))),
        (
            [np.ones((2, 5)), np.ones((2, 6))],
            {},
            TypeError,
        ),
    ],
    ids=idfn,
)
def test_check_one_sequence(x, kwargs, expects):
    if isinstance(expects, (np.ndarray, list)):
        x = check_one_sequence(x, **kwargs)
        assert_equal(x, expects)
    else:
        with pytest.raises(expects):
            x = check_one_sequence(x, **kwargs)


@pytest.mark.parametrize(
    "x,kwargs,expects",
    [
        (np.ones((1, 5)), {}, np.ones((1, 5))),
        (np.ones((5,)), {}, np.ones((1, 5))),
        (np.ones((1, 5)), {"expected_dim": (6,)}, ValueError),
        ("foo", {}, TypeError),
        (1, {}, np.ones((1, 1))),
        (
            [np.ones((1, 5)), np.ones((1, 6))],
            {},
            [np.ones((1, 5)), np.ones((1, 6))],
        ),
        (
            [np.ones((5,)), np.ones((6,))],
            {},
            [np.ones((1, 5)), np.ones((1, 6))],
        ),
        (
            [np.ones((1, 5)), np.ones((1, 6))],
            {"expected_dim": (5, 6)},
            [np.ones((1, 5)), np.ones((1, 6))],
        ),
        (
            [np.ones((1, 5)), np.ones((1, 6))],
            {"expected_dim": (6, 5)},
            ValueError,
        ),
        ([np.ones((1, 5)), np.ones((1, 6))], {"expected_dim": (5,)}, ValueError),
        ([np.ones((1, 5)), np.ones((1, 6))], {"expected_dim": (5, 8)}, ValueError),
        (["foo", np.ones((1, 6))], {}, TypeError),
        ([1, np.ones((1, 6))], {}, [np.ones((1, 1)), np.ones((1, 6))]),
        (np.ones((5, 5)), {}, np.ones((5, 5))),
        (np.ones((2, 5)), {}, np.ones((2, 5))),
        (np.ones((2, 5)), {"expected_dim": (5,)}, np.ones((2, 5))),
        (np.ones((2, 5)), {"expected_dim": (2,)}, ValueError),
        (
            [np.ones((2, 5)), np.ones((2, 6))],
            {},
            [np.ones((2, 5)), np.ones((2, 6))],
        ),
        (
            [np.ones((5, 5)), np.ones((5, 6))],
            {"expected_dim": (5, 6)},
            [np.ones((5, 5)), np.ones((5, 6))],
        ),
        ([np.ones((5, 5)), np.ones((5, 6))], {"expected_dim": (5,)}, ValueError),
        ([np.ones((8, 5)), np.ones((8, 6))], {"expected_dim": (5, 8)}, ValueError),
        ([np.ones((7, 5)), np.ones((8, 6))], {"expected_dim": (5, 6)}, ValueError),
        (
            [np.ones((2, 7, 5)), np.ones((2, 8))],
            {"expected_dim": ((7, 5), 8)},
            [np.ones((2, 7, 5)), np.ones((2, 8))],
        ),
        ([np.ones((2, 5)) for _ in range(5)], {}, [np.ones((2, 5)) for _ in range(5)]),
        ([np.ones((5,)) for _ in range(5)], {}, [np.ones((1, 5)) for _ in range(5)]),
        (
            [np.ones((2, 5, 7)) for _ in range(5)],
            {"expected_dim": ((5, 7),)},
            [np.ones((2, 5, 7)) for _ in range(5)],
        ),
        (
            [np.ones((5, 2, 7)), np.ones((5, 2, 6))],
            {"expected_dim": (7, 6)},
            [np.ones((5, 2, 7)), np.ones((5, 2, 6))],
        ),
        (
            [
                [np.ones((i, 7)) for i in range(10)],
                [np.ones((i, 6)) for i in range(10)],
            ],
            {},
            [
                [np.ones((i, 7)) for i in range(10)],
                [np.ones((i, 6)) for i in range(10)],
            ],
        ),
        (
            [
                [np.ones((i, 7)) for i in range(10)],
                [np.ones((i, 6)) for i in range(10)],
            ],
            {"expected_dim": (7, 6)},
            [
                [np.ones((i, 7)) for i in range(10)],
                [np.ones((i, 6)) for i in range(10)],
            ],
        ),
        (np.ones((5, 7, 6)), {"expected_dim": (7, 6)}, ValueError),
    ],
    ids=idfn,
)
def test_check_n_sequences(x, kwargs, expects):
    if isinstance(expects, (np.ndarray, list)):
        x = check_n_sequences(x, **kwargs)
        assert_equal(x, expects)
    else:
        with pytest.raises(expects):
            x = check_n_sequences(x, **kwargs)


@pytest.mark.parametrize(
    "caller,x,y,kwargs,expects",
    [
        (PlusNode(), np.ones((1, 5)), None, {}, (np.ones((1, 5)), None)),
        (PlusNode(), np.ones((5,)), None, {}, (np.ones((1, 5)), None)),
        (PlusNode(), np.ones((1, 5)), None, {"input_dim": (6,)}, ValueError),
        (PlusNode(input_dim=6), np.ones((1, 5)), None, {}, ValueError),
        (
            PlusNode(),
            np.ones((1, 5)),
            np.ones((1, 5)),
            {},
            (np.ones((1, 5)), np.ones((1, 5))),
        ),
        (
            PlusNode(),
            np.ones((5,)),
            np.ones((6,)),
            {},
            (np.ones((1, 5)), np.ones((1, 6))),
        ),
        (
            Offline(),
            np.ones((1, 5)),
            np.ones((1, 6)),
            {"output_dim": (7,)},
            ValueError,
        ),
        (PlusNode(output_dim=7), np.ones((1, 5)), np.ones((1, 6)), {}, ValueError),
        (
            PlusNode(name="plus0") & MinusNode(name="minus0"),
            {"plus0": np.ones((1, 5)), "minus0": np.ones((1, 6))},
            None,
            {},
            ({"plus0": np.ones((1, 5)), "minus0": np.ones((1, 6))}, None),
        ),
        (
            PlusNode(name="plus1") & MinusNode(name="minus1"),
            np.ones((1, 5)),
            None,
            {},
            ({"plus1": np.ones((1, 5)), "minus1": np.ones((1, 5))}, None),
        ),
        (
            PlusNode(name="plus2") & MinusNode(name="minus2"),
            {"plus2": np.ones((1, 5))},
            None,
            {},
            ValueError,
        ),
        (OnlineNode(output_dim=6), np.ones((1, 5)), np.ones((1, 5)), {}, ValueError),
        (
            OnlineNode(),
            np.ones((1, 5)),
            PlusNode(input_dim=5, output_dim=5).initialize(),
            {},
            (np.ones((1, 5)), None),
        ),
        (OnlineNode(), PlusNode(), np.ones((1, 5)), {}, TypeError),
        (Offline(), np.ones((1, 5)), PlusNode(), {}, TypeError),
        (MinusNode(), PlusNode(), np.ones((1, 5)), {}, TypeError),
        (
            OnlineNode(output_dim=6),
            np.ones((1, 5)),
            PlusNode(input_dim=5, output_dim=5).initialize(),
            {},
            ValueError,
        ),
        (
            PlusNode(name="plus3") & MinusNode(name="minus3"),
            {"plus3": np.ones((1, 5)), "minus3": np.ones((1, 6))},
            np.ones((1, 5)),
            {},
            ({"plus3": np.ones((1, 5)), "minus3": np.ones((1, 6))}, None),
        ),
        (
            PlusNode(name="plus4") >> OnlineNode(name="online4"),
            np.ones((1, 5)),
            PlusNode(),
            {},
            ({"plus4": np.ones((1, 5))}, None),
        ),
        (
            PlusNode(name="plus5") >> OnlineNode(name="online51")
            & MinusNode(name="minus5") >> OnlineNode(name="online52"),
            {"plus5": np.ones((1, 5)), "minus5": np.ones((1, 5))},
            {"online51": PlusNode(), "online52": np.ones((1, 5))},
            {},
            (
                {"plus5": np.ones((1, 5)), "minus5": np.ones((1, 5))},
                {"online52": np.ones((1, 5))},
            ),
        ),
        (
            PlusNode(name="plus6") >> OnlineNode(name="online61")
            & MinusNode(name="minus6") >> Offline(name="offline62"),
            {"plus6": np.ones((1, 5)), "minus6": np.ones((1, 5))},
            {"offline62": PlusNode(), "online61": np.ones((1, 5))},
            {},
            TypeError,
        ),
        (
            PlusNode(name="plus8") >> OnlineNode(name="online81")
            & MinusNode(name="minus8") >> Offline(name="offline82"),
            {"plus8": PlusNode(), "minus8": np.ones((1, 5))},
            {"offline82": PlusNode(), "online81": np.ones((1, 5))},
            {},
            TypeError,
        ),
        (
            PlusNode(name="plus7")
            >> Offline().fit(np.ones((10, 5)), np.ones((10, 5)))
            >> Offline(name="off7"),
            np.ones((1, 5)),
            {"off7": np.ones((1, 5))},
            {},
            ({"plus7": np.ones((1, 5))}, {"off7": np.ones((1, 5))}),
        ),
    ],
    ids=idfn,
)
def test_check_xy(caller, x, y, kwargs, expects):
    if isinstance(expects, (np.ndarray, list, dict, tuple)):
        x, y = check_xy(caller, x, y, **kwargs)
        assert_equal(x, expects[0])
        if y is not None:
            assert_equal(y, expects[1])
    else:
        with pytest.raises(expects):
            x = check_xy(caller, x, y, **kwargs)


def test_distant_feedback(plus_node, feedback_node):

    sender = PlusNode(input_dim=5, output_dim=5)
    fb = DistantFeedback(sender, feedback_node)

    fb.initialize()

    assert sender.is_initialized
    assert_equal(sender.state_proxy(), fb())

    fb = DistantFeedback(plus_node, feedback_node)

    with pytest.raises(RuntimeError):
        fb.initialize()

    plus = PlusNode(input_dim=5, output_dim=5) >> Inverter()
    minus = MinusNode(output_dim=5)
    sender = plus >> minus
    fb = DistantFeedback(sender, feedback_node)

    fb.initialize()

    assert sender.is_initialized
    assert_equal(minus.state_proxy(), fb())

    fb = DistantFeedback(plus_node, feedback_node)

    with pytest.raises(RuntimeError):
        fb.initialize()

    plus = PlusNode()
    minus = MinusNode()
    sender = plus >> minus
    fb = DistantFeedback(sender, feedback_node)

    with pytest.raises(RuntimeError):
        fb.initialize()
